

import pygame
import time
import numpy as np
import pandas as pd
import ffmpeg
import imageio
from pathlib import Path
from tqdm import tqdm_notebook
from ipywidgets import interact,widgets
from IPython.display import display 
from PIL import Image

class Simulation:

    def __init__(self,env,fps = 10,name = None):

        self.env = env
        self.fps = fps
        self._reset_frame_cache()

        name = self.__class__.__name__ if name is None else name
        self.name = f"{name}_{int(time.time())}"
    


    def _reset_frame_cache(self):
        self.frame_cache = []

    def cache_frame(self):
        frame = self.env.get_frame()
        self.frame_cache.append(frame)


    def step(self):
        """Method that can to be overridden by children simulations
        This will be run during the main run method
        TODO not forced to render
        """

        # Step for the environment
        reward,done,episode_data = self.env.step()

        # Render all objects in the environment
        self.env.prerender()
        self.env.render()
        self.env.postrender()

        return reward,done,episode_data



    def save_simulation_video(self,save,frames = None,filepath = None):

        # Saving loop
        if save is not None:
            if save == True:
                filepath = None
            elif save == False:
                return None
            elif isinstance(save,str):
                filepath = save
            else:
                raise SimulationError("Save must be True or a string to name the simulation gif")
                    
            if frames is None:
                frames = self.frame_cache

            if filepath is None:

                filepath = f"{self.name}.mp4"

            img_dir = Path("./captures")
            img_dir.mkdir(exist_ok=True)
            filepath = f"./captures/{filepath}"

            print(f"[INFO] Saving video at {filepath}")
            self._save_video_from_images(filepath,frames,fps = self.fps)


    @staticmethod
    def _save_video_from_images(filepath,images, fps=60, vcodec='libx264'):
        # From https://github.com/kkroening/ffmpeg-python/issues/246
        if not isinstance(images, np.ndarray):
            images = np.asarray(images)
        n,height,width,channels = images.shape
        process = (
            ffmpeg
                .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(width, height))
                .output(filepath, pix_fmt='yuv420p', vcodec=vcodec, r=fps)
                .overwrite_output()
                .run_async(pipe_stdin=True)
        )
        for frame in images:
            process.stdin.write(
                frame
                    .astype(np.uint8)
                    .tobytes()
            )
        process.stdin.close()
        process.wait()




    def save_simulation_gif(self,save,frames = None,filepath = None):

        # Saving loop
        if save is not None:
            if save == True:
                filepath = None
            elif save == False:
                return None
            elif isinstance(save,str):
                filepath = save
            else:
                raise SimulationError("Save must be True or a string to name the simulation gif")
                    
            if frames is None:
                frames = self.frame_cache

            if filepath is None:

                filepath = f"{self.name}.gif"

            img_dir = Path("./captures")
            img_dir.mkdir(exist_ok=True)
            filepath = f"./captures/{filepath}"

            print(f"[INFO] Saving gif at {filepath}")
            imageio.mimsave(filepath,frames,fps=self.fps)



    def replay_episode(self,fps = 5):

        # Prepare widgets
        play = widgets.Play(
            value=0,
            min=0,
            max=len(self.frame_cache) - 1,
            step=1,
            interval=int(1000/fps),
            description="Press play",
            disabled=False
        )
        slider = widgets.IntSlider(min = 0,value = 0,max = len(self.frame_cache) - 1,step = 1)
        widgets.jslink((play, 'value'), (slider, 'value'))

        # Visualize frames and widgets
        @interact(i = play)
        def show(i):
            img = Image.fromarray(self.frame_cache[i])
            return img

        display(slider)


    def run_episode(self,n_steps = 100,save = None,replay = False,fps_replay = 5,save_format = "video",render = True):


        # Simulation variables
        simulation_on = True
        i = 0
        clock = pygame.time.Clock()
        episode_rewards = []
        episode_data = []

        # Create progress bar
        progress_bar = tqdm_notebook(total=n_steps)

        # Cache first frame if needed
        if render:
            if save is not None or replay:
                self.cache_frame()

        # Set data for first initialization in case it was not computed
        # TODO optimize and remove this requirement
        self.env.set_data()

        # Main simulation loop for one episode
        while simulation_on:

            # Main step function
            reward,done,step_data = self.step()
            episode_rewards.append(reward)
            episode_data.extend(step_data)

            if render:
                # Cache frame if needed
                if save is not None or replay:
                    self.cache_frame()

                # Wait between frames
                clock.tick(self.fps)

            # Quitting loop
            for event in pygame.event.get():

                # Launch functions that can be overridden
                self.on_event(event)

                # Quit if click on QUIT
                if event.type == pygame.QUIT:
                    simulation_on = False

            # Stopping conditions
            if i >= n_steps - 1:
                simulation_on = False
            else:
                i += 1

            # Update progress bar
            progress_bar.update(1)

            # Stop simulation if done is signaled in the inner loop
            if done:
                simulation_on = False


        # Saving simulation as gif or video
        if render:
            if save_format == "video":
                self.save_simulation_video(save = save)
            elif save_format == "gif":
                self.save_simulation_gif(save = save)
            else:
                raise Exception("save_format must be 'video' or 'gif'")
        progress_bar.close()

        # Quit simulation
        self.env.quit()

        # Replay
        if render:
            if replay:
                self.replay_episode(fps = fps_replay)

        # Return episode reward
        episode_reward = np.sum(episode_rewards)

        # Return episode data
        episode_data = pd.DataFrame(episode_data)
        return episode_reward,episode_data


    def get_mouse_pos(self):
        # TODO could be in the environment class
        x,y = pygame.mouse.get_pos()
        x = x // self.env.cell_size
        y = y // self.env.cell_size
        return x,y 

    def event_is_click(self,event):
        LEFT = 1
        return event.type == pygame.MOUSEBUTTONDOWN and event.button == LEFT

    def event_is_rightclick(self,event):
        RIGHT = 3
        return event.type == pygame.MOUSEBUTTONDOWN and event.button == RIGHT

    def on_event(self,event):
        pass



class SimulationError(Exception):
    pass




from collections import defaultdict

import datetime
import pandas as pd
import json
import matplotlib.pyplot as plt
import plotly.io as pio
import visdom
import webbrowser

plt.style.use("bmh")
plt.rcParams['axes.facecolor'] = 'white'
plt.rcParams['axes.grid'] = False
plt.rcParams['axes.spines.right'] = False
plt.rcParams['axes.spines.top'] = False



class Logger:

    def __init__(self,use_visdom = False,graphs = None,freq_update = 1):
        
        self.data = defaultdict(list)
        self.use_visdom = use_visdom
        self.graphs = graphs
        self._clock_ref = None
        self._clock = 0
        self._freq_update = freq_update

        if use_visdom:
            self.init_visdom_server()


    @property
    def df(self):
        return pd.DataFrame(self.data)


    def init_visdom_server(self):
        self.visdom_server = visdom.Visdom()

    def __getitem__(self,key):
        return self.data[key]

    @staticmethod
    def open_visdom_interface():
        webbrowser.open('http://localhost:8097')


    def log_metric(self,name,value = None):

        if isinstance(name,dict):
            for k,v in name.items():
                self.log_value(k,v)
        else:

            if self.use_visdom:

                if self._clock_ref is None:
                    self._clock_ref = name

                if name == self._clock_ref:
                    self._clock += 1

                if self._clock % self._freq_update == 0:
                    self.send_visdom_server()

            
            self.data[name].append(value)


    # def make_data_plotly(self,cols = None,data = None):
    #     if cols is None: cols = list(self.data.keys())
    #     if not isinstance(cols,list): cols = [cols]
    #     if data is None:
    #         data = pd.DataFrame.from_dict(self.data,orient = "index").T.dropna()
    #     return data[cols].melt()


    # def show_plotly(self,cols,data = None,kind = "line",return_fig = False,return_dict = False):
    #     if not isinstance(cols,list): cols = [cols]

    #     if kind == "line":
    #         data = self.make_data_plotly(cols,data)
    #         fig = px.line(data,y = "value",color = "variable")
    #         if return_fig:
    #             return fig
    #         elif return_dict:
    #             return json.loads(fig.to_json())
    #         else:
    #             fig.show()

    @staticmethod
    def _make_line_trace(y,name = "trace"):
        if len(y) == 0:
            raise Exception("Empty sequence to plot, you are probably request the wrong key from the data defaultdict")
        x = list(range(len(y)))
        y = list(y)
        trace = dict(
            x=x, 
            y=y,
            mode="lines", 
            name=name,
            type="scatter",
        )
        return trace

    @staticmethod
    def _make_layout_dict(title,xaxis = "x",yaxis = "y"):
        layout = dict(title=title, xaxis={'title': xaxis}, yaxis={'title': yaxis})
        return layout


    def make_line_chart(self,cols = None,title = None,return_dict = False,return_visdom = False):
        # Remark, plotly express is too slow we have to use the original API 

        if cols is None: cols = list(self.data.keys())
        if not isinstance(cols,list): cols = [cols]

        if title is None:
            title = f"Evolution of metrics: {', '.join(cols)}"

        traces = [self._make_line_trace(self.data[col],col) for col in cols]
        layout = self._make_layout_dict(title,"Step","Value")
        data = {"data":traces,"layout":layout,"win":title}


        if return_visdom:
            self.send_visdom_data(data)
        elif return_dict:
            return data
        else:
            return pio.show(data)


    def make_chart(self,cols,title,kind = "line",return_dict = False,return_visdom = False):

        if kind == "line":
            return self.make_line_chart(cols,title,return_dict,return_visdom)
        else:
            raise Exception("Not Implemented Error")


    def send_visdom_data(self,data):
        self.visdom_server._send(data)



    def send_visdom_server(self):

        if self.graphs is None:
            for name in self.data:
                self.make_line_chart(name,return_visdom = True)

        else:
            for title,graph in self.graphs.items():
                self.make_chart(graph["cols"],title,graph["kind"],return_visdom = True)





    def save(self):
        pass


    def show(self,key):

        pd.Series(self[key]).plot(figsize = (15,4),title = f"Westworld logging for metric {key}")
        plt.show()







from copy import deepcopy
import pandas as pd
import numpy as np
import itertools

class Environment:

    def __init__(self,agents):

        if isinstance(agents,list):
            self._agents = {agent.id:agent for agent in agents}
        else:
            self._agents = agents

    @property
    def agents(self):
        return list(self._agents.values())

    @property
    def ids(self):
        return iter(deepcopy(list(self._agents.keys())))
    

    def __iter__(self):
        return iter(self._agents.values())


    def __getitem__(self,key):
        return self.get(key)


    def loc(self,key):
        return self._agents.get(key)

    def get(self,key,as_array = True):
        if as_array:
            return np.array([x[key] for x in self.agents])
        else:
            series = {x.id:x[key] for x in self.agents}
            return pd.Series(series)

    def to_df(self):
        agents = [x.to_dict() for x in self.agents]
        return pd.DataFrame(agents).set_index("id")

    def inverse_loc(self,key):
        for agent in self:
            if agent.id != key and agent.alive:
                yield agent

    def remove(self,key):
        self._agents.pop(key)


    def destroy(self):
        for _id in self.ids:
            agent = self.loc(_id)
            if agent.destroyed:
                self.remove(_id)


    def step(self):
        for agent in self.agents:

            # Equivalent to agent?
            if agent.alive:
                agent.step(self)

        self.destroy()


    def interactions(self):
        combinations = itertools.combinations(self.ids,2)
        interactions = []
        for agent1,agent2 in combinations:
            agent1,agent2 = self.loc(agent1),self.loc(agent2)
            inter_comb,interaction = agent1.interacts_with(agent2)
            if inter_comb:
                interactions.append((agent1,agent2,interaction))
        return interactions





import uuid
import pygame
from pygame.sprite import Sprite

from ..exceptions import ObjectNotBoundError



class BaseObject(Sprite):
    """Default Object serving as base for each agent evolving in environments
    """
    def __init__(self):

        # Init base sprite class
        super().__init__()

        # Create id and clock
        self.id = str(uuid.uuid1())
        self._clock = 0

        # Init function to be overridden
        self.init()


    @property
    def clock(self):
        """Default internal clock for each object
        """
        return self._clock

    def clocktick(self):
        """Helper function to increment internal clock
        """
        self._clock += 1


    def kill(self):
        if self.id in self.env._objects.keys():
            self.env.remove_object(self)


    @property
    def env(self):
        """Environment linked to the object
        """
        if self.is_bound():
            return self._env
        else:
            raise ObjectNotBoundError("The object must be attached to an environment")


    def is_bound(self):
        return hasattr(self,"_env")


    def bind(self,env):
        """Default setter function
        """
        self._env = env
        self.post_bind()

    def post_bind(self):
        pass


    def init(self):
        """Default init function 
        Can be overriden in special objects
        """
        pass


    @property
    def on_grid(self):
        return self.env.is_grid

    @property
    def stationary(self):
        return True

    @property
    def blocking(self):
        return True

    @property
    def trigger(self):
        return False

    @property
    def collectible(self):
        return False

    @property
    def radius(self):
        return 1



    @property
    def layer(self):
        return False


    @property
    def obstacle(self):
        return self.stationary and self.blocking



    def collides(self) -> list:
        """Returns if an object collides with any objects in the environment
        Needs to be binded to an environment of course
        The function will computes collisions against:
            - Rectangular or circle collisions with the blocking sprite group
            - Mask collisions with the layer sprite group
            - Rectangular collisions with the trigger sprite group

        TODO:
        Could be improved to account for trigger layers

        Returns:
            list: List of objects in collisions, empty if no collision
        """
        

        collisions = []

        # Find all rect collisions
        if self.env.has_blocking():
            c = self.collides_group(self.env.group_blocking,method = self.collision_method)
            collisions.extend(c)

        # Find all mask collisions
        if self.env.has_layers():
            c = self.collides_group(self.env.group_layers,method = "mask")
            collisions.extend(c)

        if self.env.has_triggers():
            triggers = self.collides_group(self.env.group_triggers,method = "rect")
            for trigger in triggers:
                trigger.on_collision(self)

        return collisions




    def collides_group(self,group:pygame.sprite.Group, method:str = "rect", ratio:float = None) -> list:
        """Collision helper functions, computes if an object enters a collision with a sprite group
        Three methods are available (rect, circle and mask)

        Args:
            group (pygame.sprite.Group): The target sprite group on which to test collisions
            method (str, optional): Collision method, either rectangular, circular or using a pixel perfect mask. Defaults to "rect".
            ratio (float, optional): Apply a ratio around the center of the object, useful to get collisions in a circle range. Defaults to None.

        Returns:
            list: List of objects in collisions, empty if no collision
        """

        # Verify method is in acceptable values
        assert method in ["rect","circle","mask"]

        # Rectangular collider
        if method == "rect":
            if ratio is None:
                collisions = pygame.sprite.spritecollide(self,group,False)
            else:
                collisions = pygame.sprite.spritecollide(self,group,False,pygame.sprite.collide_rect_ratio(ratio))

        # Circle collider
        elif method == "circle":
            # Warning, to check ratio works with rect but for circle 
            # We need to add the cell_size, maybe works as a radius and not a ratio

            if ratio is  None:
                collisions = pygame.sprite.spritecollide(self,group,False,pygame.sprite.collide_circle)
            else:
                ratio = ratio * self.env.cell_size
                collisions = pygame.sprite.spritecollide(self,group,False,pygame.sprite.collide_circle_ratio(ratio))

        # Mask collider
        elif method == "mask":
            collisions = pygame.sprite.spritecollide(self,group,False,pygame.sprite.collide_mask)

        # Filter to remove self object from collisions 
        collisions = [x for x in collisions if x.id != self.id]

        return collisions


    def render_circle(self,x = None,y = None,radius = None,color = None,screen = None,thickness = 0,cell_size = None):

        # Default arguments
        if x is None: x = self.x
        if y is None: y = self.y
        if color is None: color = self.color
        if screen is None: screen = self.env.screen
        if cell_size is None: cell_size = self.env.cell_size
        if radius is None: radius = self.radius 

        # When in a grid, the center is normalized on cell size
        if cell_size is not None:
            x = int((x + 0.5) * cell_size)
            y = int((y + 0.5) * cell_size)
            radius = int(radius * cell_size)

        # Draw a circle on the grid using pygame
        pygame.draw.circle(screen,color,(x,y),radius,thickness)

    
    def render_image(self):
        pass

    
    def render_text(self,text,font = "Arial",size = 30,screen = None,color = (255,255,255)):
        
        x = self.x
        y = self.y
        cell_size = self.env.cell_size
        
        position = (
            int(x* cell_size),
            int(y* cell_size),
        )

        self.env.render_text(text,font,size,position,screen,color = color)



    def render_rect(self,x = None,y = None,width = None,height = None,radius = None,color = None,center = False,screen = None,cell_size = None,thickness = 0):

        # Default arguments
        if x is None: x = self.x
        if y is None: y = self.y
        if color is None: color = self.color
        if screen is None: screen = self.env.screen
        if cell_size is None: cell_size = self.env.cell_size
        if width is None: width = self.width
        if height is None: height = self.height
        if radius is not None:
            width = radius * 2
            height = radius * 2

        # Center the square on x,y or not
        if not center:
            dimensions = (x*cell_size,y*cell_size,width * cell_size,height*cell_size)
        else:
            dimensions = (
                int((x + 0.5 - 0.5 * width) * cell_size),
                int((y + 0.5 - 0.5 * height) * cell_size),
                width * cell_size,
                height * cell_size,
            )


        # Draw rectangle using pygame
        pygame.draw.rect(screen,color,dimensions,thickness)


    def render_img(self,image = None,x = None,y = None,screen = None,cell_size = None):

        if x is None: x = self.x 
        if y is None: y = self.y
        if cell_size is None: cell_size = self.env.cell_size
        if image is None: image = self.image
        if screen is None: screen = self.env.screen
        pos = (x * cell_size,y * cell_size)
        screen.blit(image,pos)




import numpy as np
import pandas as pd
import pygame
import random
from PIL import Image

try:
    from win32api import GetSystemMetrics
except:
    GetSystemMetrics = lambda x: x


# Custom libraries
from ..algorithms.neighbors import NeighborsFinder


class GridEnvironment:
    def __init__(self,width = None,height = None,cell_size = 10,
        objects = None,show_grid = False,grid_color = (50,50,50),background_color = (0,0,0),
        callbacks_step = None,max_spawn_trials = 1000,
        toroidal = True,
        ):

        # Init pygame
        pygame.init()
        self.screen = pygame.display.set_mode((1,1),0,0)

        # Store parameters
        self._clock = 0
        self._cell_size = cell_size
        self._done = False
        self.show_grid = show_grid
        self.grid_color = grid_color
        self.max_spawn_trials = max_spawn_trials
        self.width = width if width is not None else ((GetSystemMetrics(0) - 100) // cell_size)
        self.height = height if height is not None else ((GetSystemMetrics(1) - 100) // cell_size)
        self.background_color = background_color
        self.callbacks_step = [] if callbacks_step is None else callbacks_step
        self.toroidal = toroidal

        # Groups initialization
        self.group_blocking = pygame.sprite.Group()
        self.group_layers = pygame.sprite.Group()
        self.group_triggers = pygame.sprite.Group()

        # Objects initialization
        self._objects = {}
        self.add_object(objects)

        # Grid Environment parameters
        self.setup_screen()

        # Init objects data
        self.set_data()
        self.render()

        # Prepare logs
        self._logs = []
        


    @property
    def is_grid(self):
        return self.cell_size > 1


    def make_group(self,condition = None):
        group = pygame.sprite.Group()

        if condition is not None:
            objs = self.find_objects(condition = condition,return_objects = True)
            group.add(objs)

        return group



    @property
    def cell_size(self):
        return self._cell_size

    def init_screen(self):
        pass



    @property
    def layers(self):
        return self.group_layers.sprites()

    def has_layers(self):
        return len(self.group_layers) > 0
    
    def has_blocking(self):
        return len(self.group_blocking) > 0

    def has_triggers(self):
        return len(self.group_triggers) > 0


    def quit(self):
        pygame.quit()


    def finish_episode(self):
        self._done = True


    @property
    def clock(self):
        """Default internal clock for each object
        """
        return self._clock

    def clocktick(self):
        """Helper function to increment internal clock
        """
        self._clock += 1

    @property
    def data(self):
        return self._data

    @property
    def done(self):
        return self._done


    @property
    def objects(self):
        return list(self._objects.values())


    def get_object(self,object_id):
        if isinstance(object_id,str):
            return self._objects[object_id]
        else:
            return self._objects[object_id.id]


    def __getitem__(self,object_id):
        return self.get_object(object_id)

    def remove_object(self,object_id):

        if isinstance(object_id,str):
            obj = self._objects.pop(object_id)
        else:
            obj = self._objects.pop(object_id.id)

        self.group_blocking.remove(obj)
        self.group_triggers.remove(obj)
        self.group_layers.remove(obj)



    def add_object(self,obj):

        if obj is not None:

            # If we add a list of objects, using recursive function to add each object
            if isinstance(obj,list):
                for o in obj:
                    self.add_object(o)
                
            # Add object 
            else:
                
                # Set environment as attribute for easy manipulation
                obj.bind(self)
                self._objects[obj.id] = obj

                if obj.layer:
                    self.group_layers.add(obj)

                if obj.blocking:
                    self.group_blocking.add(obj)

                if obj.trigger:
                    self.group_triggers.add(obj)
                


    def find(self,return_pos = False,return_objects = True,return_data = False,**kwargs):
        # TODO is it faster with numpy or pandas
        # The loop could be accelerated with numba?
        # Previously called find_objects
        ids = []

        if len(kwargs) == 0:
            ids = [x.id for x in self.objects]
        else:
            def helper_check_fn(obj,k,v):
                if not hasattr(obj,k):
                    return False
                else:
                    return getattr(obj,k) == v

            for obj in self.objects:
                if all([helper_check_fn(obj,k,v) for k,v in kwargs.items()]):
                    ids.append(obj.id)
                

        if return_pos:
            return [self._objects[k].pos for k in ids]
        elif return_data:
            return self.data.loc[ids]
        elif not return_objects:
            return ids
        else:
            return [self._objects[k] for k in ids]







    #=================================================================================
    # COLLIDERS
    #=================================================================================


    def wrap_in_toroidal_envs(self,x,y):

        if self.toroidal:

            env_width = self.width
            env_height = self.height

            # Check with x
            if x >= env_width:
                new_x = 0
            elif x < 0:
                new_x = env_width - 1
            else:
                new_x = x

            # Check with y
            if y >= env_height: 
                new_y = 0
            elif y < 0:
                new_y = env_height - 1
            else:
                new_y = y

            return new_x,new_y

        else:
            return x,y


    def is_object_colliding(self,obj):
        is_collision,_ = obj.collides()
        return is_collision


    #=================================================================================
    # SPAWNERS
    #=================================================================================



    def spawn(self,spawner,n,allow_overlap = False,**kwargs):
        """Spawner function
        """

        # Spawn n elements (works also with n = 1)
        for i in range(n):

            if not self.is_grid and allow_overlap == False:

                x,y = self.get_random_available_pos(allow_overlap = True)
                obj = spawner(x,y,**kwargs)
                self.add_object(obj)
                is_spawned = not obj.collides()[0]
                i = 0

                while not is_spawned:
                    x,y = self.get_random_available_pos(allow_overlap = True)
                    obj.move_at(x = x,y = y)
                    is_spawned = not obj.collides()[0]
                    
                    i += 1
                    if i > self.max_spawn_trials:
                        is_spawned = True
                        self.remove_object(obj)

            else:

                # Generate a random position considering or not overlaps
                x,y = self.get_random_available_pos(allow_overlap = allow_overlap)

                # Spawn new object using spawner
                # Pass also kwargs to spawner
                obj = spawner(x,y,**kwargs)

                self.add_object(obj)

        # Append to data
        self.set_data()




    #=================================================================================
    # MESHES AND ARRAYS
    #=================================================================================


    def get_grid(self):
        return np.zeros((self.width,self.height)).T


    def get_navigation_mesh(self,obj = None):

        # Prepare empty mesh        
        mesh = self.get_grid()

        # Get all blocking objects
        object_id = "" if obj is None else obj.id
        positions = [x.pos_array for x in self.objects if (x.id != object_id and x.blocking)]
        positions = [y for x in positions for y in x]

        # Update mesh with blocking positions
        if len(positions) > 0:
            mesh[tuple(np.array(positions).T)] = 1


        # Update navigation mesh with obstacles layer if any
        if self.has_layers:
            all_meshes = [mesh]
            all_meshes.extend([layer.get_navigation_mesh() for layer in self.layers])
            mesh = np.int8(np.sum(all_meshes,axis = 0) > 0)

        return mesh

    def get_available_mesh(self):

        # Prepare empty mesh        
        mesh = self.get_grid()

        # Get all blocking objects
        positions = [y for x in self.objects for y in x.pos_array]

        if len(positions) > 0:

            # Update mesh with blocking positions
            mesh[tuple(np.array(positions).T)] = 1

        return mesh


    def get_available_pos(self):

        # Get mesh
        mesh = self.get_available_mesh()

        # Return list of pos
        return np.array(np.where(mesh == 0)).T


    def get_random_available_pos(self,allow_overlap = False):

        if allow_overlap:
            x = np.random.randint(0,self.width)
            y = np.random.randint(0,self.height)
        else:
            # Warning x,y are reversed in numpy
            pos = self.get_available_pos()
            y,x = random.choice(pos)
        return x,y



    #=================================================================================
    # DATA MANIPULATION
    #=================================================================================


    def _prepare_data(self):
        # TODO this function is not optimized at all, 
        # It takes a few ms for only a few agents in the environment
        # Iteration is quite fast, it's transforming to dataframe that takes time
        # Yet we can probably use dataframes for faster computing later on, maybe interesting to use faster equivalents
        # Such a numpy arrays, jax arrays or torch tensors
        data = [obj.get_data() for obj in self.objects]
        if len(data) == 0:
            return None
        else:
            data = pd.DataFrame(data).set_index("id")
            return data


    def set_data(self):

        # Update data
        self._data = self._prepare_data()

        # Update also neighbors finder
        # self.neighbors_finder = NeighborsFinder(self._data)







    #=================================================================================
    # ENV LIFECYCLE
    #=================================================================================

    def callback_step(self):
        pass

    def post_step(self):
        pass

    def compute_reward(self):
        return 0

    def step(self):

        # Iterate for each object
        for agent in self.objects:

            # Only step with non static objects: ie agents
            if agent.stationary == False:
                agent.step()
                agent.clocktick()

        # Reinitialize data
        # TODO this can be optional for performance
        # Actually it's only required for computations such as pathfinding
        self.set_data()

        # Step callback
        for fn in self.callbacks_step:
            fn(self)


        self.post_step()

        # Prepare reward and done
        # TODO add reward
        reward = self.compute_reward()
        done = self.done

        # Increment internal clock
        self.clocktick()

        # Prepare step data
        step_data = self.get_logs()
        self.reset_logs()

        return reward,done,step_data


    def log(self,d):
        d = {"step":self.clock,**d}
        self._logs.append(d)

    def get_logs(self):
        return self._logs

    def reset_logs(self):
        self._logs = []


    #=================================================================================
    # RENDERERS
    #=================================================================================




    def _init_screen_from_layer(self):

        # Init grid map from layers
        layer = self.layers[0]
        w,h = layer.get_size()
        self.width = w//self.cell_size
        self.height = h//self.cell_size




    def setup_screen(self):
        """Setup the first PyGame Screen
        """

        if self.has_layers():
            self._init_screen_from_layer()
        

        # Init pygame window
        pygame.init()
        pygame.display.set_caption("Westworld environment")

        # Create window
        self.screen = pygame.display.set_mode((
            self.width * self.cell_size,
            self.height * self.cell_size
            ),pygame.RESIZABLE)

        # Fill with black
        self.reset_screen()


    def render_text(self,text,font = "Arial",size = 30,position = (5,5),screen = None,color = (0,0,0)):

        font = pygame.font.SysFont(font, size)
        text_surface = font.render(f"{text}", False,color)

        if screen is None: screen = self.screen
        screen.blit(text_surface,position)




    def reset_screen(self):
        """Reset PyGame screen by filling with BLACK color
        """
        self.screen.fill(self.background_color)


    def render_grid(self,color = (50,50,50)):

        for x in range(self.width):
            for y in range(self.height):
                rect = pygame.Rect(x*self.cell_size, y*self.cell_size,self.cell_size, self.cell_size)
                pygame.draw.rect(self.screen, color, rect, 1)


    def prerender(self):
        pass


    def postrender(self):
        pass


    def render(self,screen = None):
        """Render the environment
        TODO: could be without PyGame
        """

        if screen is None:
            # Reset current screen to black
            self.reset_screen()
            screen = self.screen

        # Show grid if necessary
        if self.show_grid:
            self.render_grid(self.grid_color)

        # Render each object
        for el in self.objects:
            el.prerender()
            el.render(screen = screen)
            el.postrender()

        # Update the PyGame renderer
        pygame.display.update()



    def get_frame(self):
        """Get the current frame as numpy array

        Returns:
            np.ndarray: The current frame as numpy array
        """
        return pygame.surfarray.array3d(self.screen).swapaxes(0,1)


    def get_img(self):
        """Get the current frame as PIL image

        Returns:
            PIL.Image: The current frame as PIL Image
        """
        img = self.get_frame()
        return Image.fromarray(img)


    def save_img(self,filepath):
        """Save the current frame in a picture file (eg .png)

        Args:
            filepath (str): The output filename
        """
        img = self.get_img()
        img.save(filepath)


        




        

        





        





from .base_object import BaseObject
from .rectangle import BaseRectangle
from .collectible import BaseCollectible
from .trigger import BaseTrigger
from .obstacle import BaseObstacle
from .layer import BaseLayer



from collections import defaultdict
from scipy.stats import norm
import numpy as np
import random

from .agent import BaseAgent
from ..colors import *


class SIRAgent(BaseAgent):

    def __init__(self,x,y,state = "S",recovery_duration_range = [20,50],contact_risk = 3,show_search_radius = False,search_radius_method = "circle",search_radius = 2,*args,**kwargs):

        super().__init__(x,y,*args,**kwargs,search_radius = search_radius,img_asset = "blob")
        
        self.set_state(state)
        self.show_search_radius = show_search_radius
        self.search_radius_method = search_radius_method
        self.recovery_duration = np.random.randint(*recovery_duration_range)
        self.contact_risk = contact_risk
        self.infected_date = 0

        
    def set_state(self,state):
        
        self.state = state
        
        if state == "S":
            self.color = BLUE
        elif state == "I":
            self.color = RED
        else:
            self.color = GREEN
            
        self.init_internals(errors_binding = False)
        
        
    
    def step(self):
        
        if self.state == "S":
            
            n_infected = len(self.find_in_range({"state":"I"}))
            
            if n_infected > 0:
                proba_infection = norm.cdf(n_infected,loc = self.contact_risk,scale = 3)

                if random.random() < proba_infection:

                    self.set_state("I")
                    self.infected_date = self.clock
                
                
        elif self.state == "I":
            
            if self.clock - self.infected_date >= self.recovery_duration:
                self.set_state("R")
            
        self.wander()
        
        
        
        



from PIL import Image
import numpy as np
import pandas as pd
import pyperclip


def make_lighter_color(color,lighter = 0.5):
    lighter_fn = lambda x,t : min([255,int((1+t)*x)]) 
    return tuple([lighter_fn(x,lighter) for x in color])

def get_unique_colors_sprite(filepath):
    img = Image.open(filepath)
    array = np.array(img).reshape(-1,3).tolist()
    array = ["(" + ",".join(map(str,x)) + ")" for x in array]
    return pd.Series(array).value_counts()


def make_sprite_generator_fn(filepath,replace = None,return_array = False):
    
    img = Image.open(filepath)
    shape = img.size
    array = np.array(img).reshape(-1,3).tolist()
    array = ["(" + ",".join(map(str,x)) + ")" for x in array]
    
    if replace is not None:
        array = [replace.get(x,x) for x in array]
        
    array = np.array(array).reshape(shape)
    
    if return_array:
        return array
    
        
    array = f"{array.tolist()}".replace(",",",\n\t\t").replace('"','').replace("'","")
    
    fn_str = f"import numpy as np\n\ndef make_sprite(args):\n\treturn np.array(\n\t\t{array}\n)"
    fn_str = fn_str.replace("\t","    ")
    pyperclip.copy(fn_str)
    print("Copied to clipboard")




import attr

@attr.s(slots = True)
class Agent:

    # Agent id
    id = attr.ib()
    id.default
    def _init_id(self):
        return str(uuid.uuid1())


    # Inactive after destroy
    _destroyed = attr.ib(repr = False,default = False)


    def destroy(self):
        self._destroyed = True

    @property
    def destroyed(self):
        return self._destroyed

    @property
    def alive(self):
        return not self._destroyed

    def __getitem__(self,key):
        return self.get(key)


    def get(self,key):
        return getattr(self,key)

    def to_dict(self):
        return attr.asdict(self)


    def interacts_with(self,other):
        return False
    
    


from .agent import BaseAgent

class CollectibleFinderAgent(BaseAgent):


    def __init__(self,x,y,show_search_radius = True,search_radius_method = "circle",search_radius = 2,*args,**kwargs):

        super().__init__(x,y,*args,**kwargs,search_radius = search_radius)

        self.show_search_radius = show_search_radius
        self.search_radius_method = search_radius_method
        self.targets = {}
        self.current_target = None

    
    def step(self):
        
        targets = self.find_in_range(condition = {"collectible":True})
        self.targets.update({t.id:t for t in targets if t.id not in self.targets})
        self.targets = {k:v for k,v in self.targets.items() if k in self.env._objects}
        
        if len(self.targets) == 0:
            self.wander()
            
        else:
            if self.current_target is None:
                self.current_target = list(self.targets.values())[0]
                
            reached_target = self.move_towards(obj = self.current_target,n = 100)
            if reached_target:
                self.current_target = None



import numpy as np

def image3d_to_mask(img,threshold = 0.5):
    mask = 1 - np.int8((np.max(img,axis = 2) / 255 ) > threshold)
    return mask


def snap_mask_to_grid(mask,cell_size,threshold = 0.5):

    new_mask = np.copy(mask)
    width = mask.shape[0] // cell_size
    height = mask.shape[1] // cell_size

    for x in range(width):
        for y in range(height):
            index = np.s_[x*cell_size:(x+1)*cell_size,y*cell_size:(y+1)*cell_size]
            submask = mask[index]
            if submask.mean() > threshold:
                new_mask[index] = 1
            else:
                new_mask[index] = 0

    return new_mask


def mask_to_image3d(mask):
    return np.repeat((1 - mask[:,:,np.newaxis])*255,3,axis = 2)


def mesh_to_mask(mesh,cell_size):
    """Transform a numpy array to another numpy array but with each cell repeated cell_size times
    Useful when we have a navigation mesh or a maze where each value in the numpy array represents one cell
    and we want to convert to image where each value is a pixel, ie each value is repeated cell_size pixel times


    Args:
        mask (np.ndarray): A numpy array with 0 and 1 representing obstacles
        cell_size (int): the size for each square cell in the grid
    """

    w,h = mesh.shape
    new_w,new_h = w * cell_size,h*cell_size
    mask = np.zeros((new_w,new_h))

    for x in range(w):
        for y in range(h):
            index = np.s_[x*cell_size:(x+1)*cell_size,y*cell_size:(y+1)*cell_size]
            mask[index] = mesh[x,y]

    mask = mask.astype(np.int8)
    return mask



def mask_to_mesh(mask,cell_size,threshold = None):

    width = mask.shape[0] // cell_size
    height = mask.shape[1] // cell_size
    mesh = np.zeros((width,height))

    for x in range(width):
        for y in range(height):
            index = np.s_[x*cell_size:(x+1)*cell_size,y*cell_size:(y+1)*cell_size]
            submask_mean = mask[index].mean()

            if threshold is None:
                mesh[x,y] = submask_mean
            else:
                assert 0 < threshold < 1
                mesh[x,y] = int(submask_mean > threshold)

    return mesh



import pandas as pd
from scipy.spatial.kdtree import KDTree

class NeighborsFinder:
    def __init__(self,data):
        """Find neigbors
        
        Args:
            data (pd.DataFrame or dict): data with at columns id,pos,and eventually range
        """
        if data is not None:

            self.data = data

            # Safety checks
            assert isinstance(data,pd.DataFrame)
            assert "pos" in data.columns


            # Initialize KDTree finder from scipy
            self.tree = KDTree(self.data["pos"].tolist())


    def find_in_range(self,obj,search_range):

        # TODO 
        # Will not work with double colliders
        # Can be made using colliders in PyGame ?

        # Safety check
        assert hasattr(self,"tree")

        # Get position from dataset
        pos = obj.pos

        # Find neighbors in range
        # We remove the first one which is the identity object
        idx = self.tree.query_ball_point(pos,search_range)
        
        # Return filtered data
        ids = self.data.iloc[idx].index.tolist()
        assert obj.id not in ids
        return ids


    def find_closest(self,obj,k = 1):

        # Safety check
        assert hasattr(self,"tree")

        # Get object position from which we want to find neighbors
        pos = obj.pos

        # Get position from dataset
        distances,idx = self.tree.query(pos,k = k)

        if k == 1:
            distances = [distances]
            idx = [idx]

        # Get ids from dataset
        # Safe check object is not in neighbors, which would mean above we remove another overlapping objects with [1:]
        ids = self.data.iloc[idx].index.tolist()
        assert obj.id not in ids
        return distances,ids



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['westworld',
 'westworld._deprecated',
 'westworld.agents',
 'westworld.algorithms',
 'westworld.algorithms.pathfinding',
 'westworld.assets',
 'westworld.assets.sprites',
 'westworld.environment',
 'westworld.objects',
 'westworld.optimization',
 'westworld.simulation',
 'westworld.utils']

package_data = \
{'': ['*']}

install_requires = \
['ffmpeg-python>=0.2.0,<0.3.0',
 'imageio>=2.9.0,<3.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'pygame>=2.0.1,<3.0.0',
 'scipy>=1.7.1,<2.0.0',
 'tqdm>=4.61.2,<5.0.0']

setup_kwargs = {
    'name': 'westworld',
    'version': '0.1.3',
    'description': 'Multi-agent simulation library in Python',
    'long_description': '# Westworld\n![](docs/img/cover_hq_westworld1.jpg)\n\n## Description\n**Westworld** is a multi-agent simulation library, its goal to simulate and optimize systems and environments with multiple agents interacting. Its inspiration is drawn from Unity software and [Unity ML Agents](https://github.com/Unity-Technologies/ml-agents), adapted in Python. \n\nThe goal is to be able to simulate environments in logistics, retail, epidemiology, providing pre-coded spatial environments and communication between agents. Optimization can be included using heuristics as well as Reinforcement Learning.\n\n!!! warning "Experimental"\n    This library is extremely experimental, under active development and alpha-release\n    Don\'t expect the documentation to be up-to-date or all features to be tested\n    Please contact [us](mailto:theo.alves.da.costa@gmail.com) if you have any question\n\n*The name is of course inspired by the TV series Westworld, which is actually a gigantic multi-agent simulation system.*\n\n## Documentation\nDocumentation is available locally in docs folder or online at https://theolvs.github.io/westworld\n\n## Features\n### Current features\n- Easy creation of Grid and non-grid environments\n- Objects (Agents, Obstacles, Collectibles, Triggers)\n- Subclassing of different objects to create custom objects\n- Spawner to generate objects randomly in the environment\n- Basic rigid body system for all objects\n- Simple agent behaviors (pathfinding, wandering, random walk, fleeing, vision range)\n- Automatic maze generation\n- Layer integration to convert image to obstacle and snap it to a grid\n- Sample simulations and sample agents for classic simulations\n- Simulation visualization, replay and export (gif or video)\n\n### Roadmap features\n- More classic simulations and tutorials (boids, sugarscape)\n- Better pathfinding\n- Easy Reinforcement Learning integration with Stable Baselines\n- Other visualization functions than PyGame for web integration \n\n\n## Installation\n### Install from PyPi\nThe library is available on [PyPi](https://pypi.org/project/westworld/) via \n```\npip install westworld\n```\n\n### For developers\n- You can clone the github repo / fork and develop locally\n- Poetry is used for environment management, dependencies and publishing, after clone you can run \n\n```\n# To setup the environment\npoetry install\n\n# To run Jupyter notebook or a python console\npoetry run jupyter notebook\npoetry run python\n```\n\n## Contributors\n- [Théo Alves Da Costa](mailto:theo.alves.da.costa@gmail.com)\n\n## Javascript version\nA javascript version is being developed at https://github.com/TheoLvs/westworldjs\n\n',
    'author': 'Theo Alves Da Costa',
    'author_email': 'theo.alves.da.costa@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)




import numpy as np

def make_ball(color1 = (255, 127, 39),color2 = (255, 201, 14),transparency = (200, 191, 231)):
    return np.array(   
        [[transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color2,
        color2,
        color2,
        color2,
        color2,
        color2,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        color1,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency],
        [transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency,
        transparency]]
    )